bl_info = {
    "name": "Boolean Bevel",
    "author": "Rodinkov Ilya",
    "version": (0, 0, 2),
    "blender": (2, 75, 0),
    "location": "View3D > Tools > Boolean Bevel > Bevel",
    "description": "Adds a new Mesh Object",
    "warning": "This addon is still in development.",
    "wiki_url": "",
    "category": "Object",
}

import bpy
import addon_utils
from bpy.types import AddonPreferences


class BooleanBevelPreferences(AddonPreferences):
    bl_idname = __name__

    def draw(self, context):
        layout = self.layout
        layout.label(text="I hope that there will be settings")


class ObjectBooleanBevel(bpy.types.Operator):
    """Create the bevel on Object"""
    bl_idname = "object.boolean_bevel"
    bl_label = "Boolean Bevel"
    bl_options = {'REGISTER', 'UNDO'}

    # def draw(self, context):
    #     layout = self.layout
    #     layout.label(text="I hope that there will be settings")

    wire = bpy.props.BoolProperty(name="Show wire", default=True)

    change_operation = bpy.props.EnumProperty(name="Change Boolean Operation",
                                              items=(("False", "Not change", "Not change"),
                                                     ("UNION", "Union", "Use Union"),
                                                     ("DIFFERENCE", "Difference", "Use Difference"),
                                                     ("INTERSECT", "Intersect", "Use Intersect"),
                                                     ("SLICE", "Slice", "Use Slice")),
                                              description="Change Boolean Operation",
                                              default="False")

    change_subdivide = bpy.props.BoolProperty(name="Change subdivide", default=False)
    subdiv_a = bpy.props.IntProperty(name="Subdivision A", default=1, min=0, max=6)
    subdiv_b = bpy.props.IntProperty(name="Subdivision B", default=1, min=0, max=6)

    simplify = bpy.props.FloatProperty(name="Simplify", default=0.0, min=0.0, max=1.0, unit="ROTATION", step=100)

    # simplify = bpy.props.EnumProperty(name="Simplify",
    #                                   items=(("0", "Not Use", "Not Use"),
    #                                          ("1", "1", "0.017 radians"),
    #                                          ("2", "2", "0.035 radians"),
    #                                          ("3", "3", "0.052 radians"),
    #                                          ("4", "4", "0.07 radians"),
    #                                          ("5", "5", "0.087 radians")),
    #                                   description="Simplify Guide Curve",
    #                                   default="0")

    subdivide = bpy.props.IntProperty(name="Subdivide Patch", default=0, min=0, max=500)
    relax = bpy.props.EnumProperty(name="Relax",
                                   items=(("0", "Not Use", "Not Use"),
                                          ("1", "1", "One"),
                                          ("3", "3", "Three"),
                                          ("5", "5", "Five"),
                                          ("10", "10", "Ten"),
                                          ("25", "25", "Twenty-five")),
                                   description="Number of times the loop is relaxed",
                                   default="5")

    value_radius = bpy.props.FloatProperty(name="Cut Radius", default=0.05, min=0.0001, max=30.0, step=1)
    sides = bpy.props.IntProperty(name="Sides of circle", default=0, min=0, max=10)
    fix_curve = bpy.props.BoolProperty(name="Fix Curve Twist", default=True)
    curve_tilt = bpy.props.FloatProperty(name="Mean Tilt", default=45.0, min=-45.0, max=45.0)
    twist_mode = bpy.props.EnumProperty(name="Twist Mode",
                                        items=(("MINIMUM", "Minimum", "Use the least twist over the entire curve"),
                                               ("TANGENT", "Tangent", "Use the tangent to calculate twist")),
                                        description="The type of tilt calculation for 3D Curves",
                                        default="MINIMUM")
    preview_curve = bpy.props.BoolProperty(name="Preview Curve", default=False)

    union_path = bpy.props.EnumProperty(name="Path Operation",
                                        items=(("UNION", "Union", "Use Union"),
                                               ("DIFFERENCE", "Difference", "Use Difference")),
                                        description="Change Boolean Operation",
                                        default="UNION")

    fillet_profile = bpy.props.FloatProperty(name="Fillet Profile", default=0.7, min=-0.15, max=1.0)
    fillet_segments = bpy.props.IntProperty(name="Fillet Segments", default=5, min=0, max=30)

    def execute(self, context):
        try:
            state = addon_utils.check("mesh_looptools")
            print(state)
            if not state:
                bpy.ops.wm.addon_enable(module="mesh_looptools")
        except:
            self.report({'ERROR'}, "Loop Tools not installed.")
            return {'CANCELLED'}

        scene = context.scene
        src_obj = scene.objects.active

        have_bool = prepare_object(scene, src_obj, self.change_subdivide, self.subdiv_a, self.subdiv_b,
                                   self.change_operation)
        if have_bool[0]:
            src_obj.show_wire = self.wire
            src_obj.show_all_edges = self.wire
            if self.change_operation == "SLICE":
                src_obj.select = True
                bpy.ops.object.duplicate_move(OBJECT_OT_duplicate={"linked": False, "mode": 'TRANSLATION'})
                slice_object = scene.objects.active
                slice_object.name = src_obj.name + "_SLICE"
                slice_object.modifiers[len(slice_object.modifiers)-1].operation = "INTERSECT"
                try:
                    bpy.ops.object.modifier_apply(apply_as='DATA', modifier=have_bool[1])
                except:
                    bpy.ops.object.modifier_remove(modifier=have_bool[1])

                scene.objects.active = src_obj
                bpy.ops.object.select_all(action='DESELECT')

            try:
                bpy.ops.object.modifier_apply(apply_as='DATA', modifier=have_bool[1])
            except:
                bpy.ops.object.modifier_remove(modifier=have_bool[1])
            guide = get_guide(scene, self.simplify, self.subdivide, self.relax)
            if not guide:
                self.report({'ERROR'}, "Error on get_guide")
                return {'CANCELLED'}
            else:
                create_curve(self.value_radius, self.sides, self.fix_curve, self.curve_tilt, self.twist_mode)
                if self.preview_curve:
                    return {'FINISHED'}
                else:
                    do_boolean(scene, src_obj, self.union_path)
                    create_bevel(scene, src_obj, self.fillet_profile, self.fillet_segments)

                    if self.change_operation == "SLICE":
                        scene.objects.active = slice_object
                        bpy.ops.object.mode_set(mode='EDIT')
                        bpy.ops.mesh.select_mode(type='VERT')
                        bpy.ops.mesh.select_all(action='SELECT')
                        bpy.ops.mesh.hide(unselected=False)
                        bpy.ops.object.mode_set(mode='OBJECT')
                        do_boolean(scene, slice_object, self.union_path)
                        create_bevel(scene, slice_object, self.fillet_profile, self.fillet_segments)
                    clear_objects(scene, src_obj)
                    return {'FINISHED'}
        else:
            self.report({'ERROR'}, "Object does not have a Boolean modifier")
            return {'CANCELLED'}


class BooleanBevelPanel(bpy.types.Panel):
    bl_label = "Boolean Bevel"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'TOOLS'
    bl_category = "Tools"

    def draw(self, context):
        layout = self.layout
        split = layout.split()
        col = split.column(align=True)
        col.operator("object.boolean_bevel", text="Bevel", icon='PLUGIN')


def menu_func(self, context):
    self.layout.operator(ObjectBooleanBevel.bl_idname)


def prepare_object(scene, src_obj, change_subdivide, subdiv_a, subdiv_b, change_operation):
    modifiers = src_obj.modifiers
    boolean = False
    bool_object = False
    subsurf_boolean = False
    subsurf_object = False

    if len(modifiers) > 0:
        # проверка на наличие булена
        for modifier in modifiers:
            if modifier.type == "BOOLEAN":
                boolean = modifier
                bool_object = modifier.object
                break
    if boolean and (bool_object != False):
        # смена операции
        if change_operation != "False":
            if change_operation == "SLICE":
                boolean.operation = "DIFFERENCE"
            else:
                boolean.operation = change_operation
        # ищем последние модификаторы подразделения
        for modifier in bool_object.modifiers:
            if modifier.type == "SUBSURF":
                subsurf_object = modifier
        for modifier in modifiers:
            if modifier.type == "SUBSURF":
                subsurf_boolean = modifier
        # меняем подразделение
        if change_subdivide:
            if subsurf_boolean != False:
                subsurf_boolean.levels = subdiv_a
            if subsurf_object != False:
                subsurf_object.levels = subdiv_b
        for modifier in modifiers:
            # пропускаем фаски
            if modifier.name.find("Boolean Bevel") == -1:
                # если булен
                if modifier.type == "BOOLEAN":
                    bpy.ops.object.mode_set(mode='EDIT')
                    bpy.ops.mesh.select_mode(type='VERT')
                    bpy.ops.mesh.select_all(action='SELECT')
                    bpy.ops.mesh.hide(unselected=False)
                    bpy.ops.object.mode_set(mode='EDIT')
                    bpy.ops.object.mode_set(mode='OBJECT')
                    scene.objects.active = modifier.object
                    bpy.ops.object.mode_set(mode='EDIT')
                    bpy.ops.mesh.select_mode(type='VERT')
                    bpy.ops.mesh.reveal()
                    bpy.ops.object.mode_set(mode='OBJECT')
                    modifier.object.draw_type = 'BOUNDS'
                    scene.objects.active = src_obj
                    bpy.ops.object.select_all(action='DESELECT')
                    return [True, modifier.name]
                else:
                    # если нет то применяем модификаторы
                    try:
                        bpy.ops.object.modifier_apply(apply_as='DATA', modifier=modifier.name)
                    except:
                        bpy.ops.object.modifier_remove(modifier=modifier.name)
    return [False]


def get_guide(scene, simplify, subdivide, relax):
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.select_mode(type='EDGE')
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.region_to_loop()
    try:
        bpy.ops.mesh.duplicate_move()
        bpy.ops.mesh.separate(type='SELECTED')
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.hide(unselected=False)
        bpy.ops.object.mode_set(mode='OBJECT')
        guide = bpy.context.selected_objects[0]
        guide.name = "BOOLEAN_BEVEL_GUIDE"
    except:
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.ed.undo()
        return False
    scene.objects.active = guide
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.select_mode(type='VERT')
    bpy.ops.mesh.select_all(action='SELECT')

    if simplify > 0:
        bpy.ops.mesh.dissolve_limited(angle_limit=simplify)
    if subdivide > 0:
        bpy.ops.mesh.subdivide(number_cuts=subdivide)
    if relax != '0':
        bpy.ops.mesh.looptools_relax(iterations=relax)
    bpy.ops.object.mode_set(mode='OBJECT')
    return guide


def create_curve(value_radius, sides, fix_curve, curve_tilt, twist_mode):
    bpy.ops.object.convert(target='CURVE', keep_original=True)
    curve_cut = bpy.context.object
    curve_cut.name = "BOOLEAN_BEVEL_CURVE"
    curve_cut.data.fill_mode = 'FULL'
    curve_cut.data.bevel_depth = value_radius
    curve_cut.data.bevel_resolution = sides
    if fix_curve:
        curve_cut.data.resolution_u = 1
        curve_cut.data.twist_mode = twist_mode
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.select_all(action='SELECT')
        bpy.ops.curve.spline_type_set(type='BEZIER')
        bpy.ops.curve.handle_type_set(type='FREE_ALIGN')
        bpy.ops.curve.normals_make_consistent()
        bpy.ops.transform.tilt(value=curve_tilt)
        bpy.ops.object.mode_set(mode='OBJECT')
    else:
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.select_all(action='SELECT')
        bpy.ops.transform.tilt(value=curve_tilt)
        bpy.ops.object.mode_set(mode='OBJECT')
    return curve_cut


def do_boolean(scene, src_obj, union_path):
    try:
        bpy.ops.object.convert(target='MESH', keep_original=False)
    except:
        pass
    curve_cut = bpy.data.objects['BOOLEAN_BEVEL_CURVE']
    curve_cut.draw_type = 'BOUNDS'
    boolean_modifier = src_obj.modifiers.new(name="BooleanBevel", type='BOOLEAN')
    boolean_modifier.object = curve_cut
    boolean_modifier.operation = union_path
    bpy.ops.object.select_all(action='DESELECT')
    scene.objects.active = src_obj
    bpy.ops.object.modifier_apply(apply_as='DATA', modifier=boolean_modifier.name)

    bpy.ops.object.select_all(action='DESELECT')
    scene.objects.active = src_obj


def create_bevel(scene, src_obj, fillet_profile, fillet_segments):
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.select_mode(type='VERT')
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.region_to_loop()
    bpy.ops.mesh.select_all(action='INVERT')
    bpy.ops.object.vertex_group_add()
    bpy.ops.object.vertex_group_assign()

    bpy.ops.object.mode_set(mode='OBJECT')
    current_group = src_obj.vertex_groups.active.name

    guide = bpy.data.objects['BOOLEAN_BEVEL_GUIDE']

    shrinkwrap_modifier = src_obj.modifiers.new(name="shrinkwrap", type='SHRINKWRAP')
    shrinkwrap_modifier.wrap_method = 'NEAREST_VERTEX'
    shrinkwrap_modifier.vertex_group = current_group
    shrinkwrap_modifier.target = guide
    bpy.ops.object.modifier_apply(apply_as='DATA', modifier=shrinkwrap_modifier.name)

    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.select_all(action='DESELECT')

    bpy.ops.object.vertex_group_select()
    bpy.ops.mesh.remove_doubles()

    fillet_modifier = src_obj.modifiers.new(name="Boolean Bevel", type='BEVEL')
    fillet_modifier.offset_type = 'PERCENT'
    fillet_modifier.profile = fillet_profile
    fillet_modifier.segments = fillet_segments
    fillet_modifier.use_clamp_overlap = True
    fillet_modifier.limit_method = 'VGROUP'
    fillet_modifier.vertex_group = current_group
    fillet_modifier.width = 98
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.reveal()
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.normals_make_consistent(inside=False)
    bpy.ops.object.mode_set(mode='OBJECT')


def clear_objects(scene, src_obj):
    bpy.ops.object.select_all(action='DESELECT')
    bpy.data.objects['BOOLEAN_BEVEL_CURVE'].select = True
    bpy.data.objects['BOOLEAN_BEVEL_GUIDE'].select = True
    bpy.ops.object.delete(use_global=False)
    bpy.ops.object.select_all(action='DESELECT')
    src_obj.select = True
    scene.objects.active = src_obj


def register():
    bpy.utils.register_class(BooleanBevelPanel)
    bpy.utils.register_class(ObjectBooleanBevel)
    bpy.types.VIEW3D_MT_object.append(menu_func)
    bpy.utils.register_class(BooleanBevelPreferences)


def unregister():
    bpy.utils.unregister_class(BooleanBevelPanel)
    bpy.utils.unregister_class(ObjectBooleanBevel)
    bpy.types.VIEW3D_MT_object.remove(menu_func)
    bpy.utils.unregister_class(BooleanBevelPreferences)


if __name__ == "__main__":
    register()
